import os
import re
import tempfile
import json
import logging

from subprocess import call

from django.http import HttpResponse, HttpResponseServerError
from django.template import RequestContext
from django.shortcuts import render_to_response
from django.core.urlresolvers import reverse
from django.conf import settings
from django.utils.encoding import smart_str
from django.core.mail import send_mail

from duat.models import Feedback, Project

logger = logging.getLogger(__name__)


def post(request, project_name):
    """
    Store a posted feedback submission
    """
    project = Project.objects.get(name=project_name)

    comment = page = referrer = None
    data = request.POST.get('data',None)

    if data:
        j = json.loads(data.encode("utf-8"))
        html = j['html']
        comment = j['comment']
        referrer = j['referrer']
        page = j['url']

        # process html contents...

        # remove all scripts
        # NB - We are not attempting to comprehensively block XSS, 
        # rather simply disable any javascript from loading
        p = re.compile('<script(.*?)<\/script>', flags=(re.IGNORECASE | re.DOTALL))
        html = p.sub('',html)

        pos = html.find('<head')
        pos = html.find('>',pos)+1

        # insert base tag
        if '<base' not in html and page:
            host = page[:page.find('/',10)]
            html = html[0:pos] + "<base href='"+host+"'></base>" + html[pos:]

    if comment and html:
        feedback = Feedback(project = project,
                            comment = comment,
                            html = html,
                            user_agent = request.META.get('HTTP_USER_AGENT',''),
                            referrer = referrer,
                            page = page)
        feedback.save()

        logger.info('Saved issue %s' % feedback.id)

        generate_screenshot(feedback)

        if project.notify_admin and project.admin.email:
            try:
                url = reverse('admin:%s_%s_change' % (feedback._meta.app_label,
                                                      feedback._meta.module_name),
                              args=[feedback.id])
                url = 'http%s://%s%s' % (('s' if request.is_secure() else ''),
                                         request.META.get('HTTP_HOST'),
                                         url)
                send_mail('DUAT: feedback received',
                          'You can view this item here:\n%s' % url, 
                          'web-master@%s ' % settings.HOSTNAME, 
                          [project.admin.email],
                          fail_silently=False)
            except:
                logger.warn("Unable to notify admin")

        return HttpResponse("Thanks")
    else:
        logger.error('Not saved - missing data')
        return HttpResponseServerError('Something went wrong')


def view(request, project_name, id):
    project = Project.objects.get(name=project_name)
    feedback = Feedback.objects.get(pk=id, project=project)

    html = feedback.html
    pos = html.find('<head')
    pos = html.find('>',pos)+1

    # add our own admin script
    url = request.META.get('HTTP_HOST') + reverse('duat-admin')
    html = html[0:pos] + "<script src='//%s'></script>" % url + html[pos:]

    return HttpResponse(html)


def generate_screenshot(feedback):
    """ Pass the html onto PhantomJS to convert into an image """
    PHANTOMJS_EXECUTABLE = getattr(settings, 'PHANTOMJS_EXECUTABLE', '')

    if not PHANTOMJS_EXECUTABLE:
        logger.warn('PHANTOMJS_EXECUTABLE not set - skipping screenshot generation')
        return

    output_dir = os.path.join(settings.STATIC_ROOT, 'duat', 'screenshots')
    output_file = "%s/%s.jpg" % (output_dir, feedback.id) 

    # create a temp file to write our HTML to
    input_file = tempfile.NamedTemporaryFile(delete=True)

    input_file.write(smart_str(feedback.html))
    # insert base so that static resources are loaded correctly

    # construct parameters to our phantom instance
    args = [PHANTOMJS_EXECUTABLE,
            os.path.dirname(__file__)+"/renderhtml.js",
            input_file.name,
            output_file]

    # create the process
    p = call(args)


def generate_js(request, filename, project_name=None):
    """ 
    Returns a javascript file generated using django's templating mechanism
    """
    project = path = None
    if project_name:
        project = Project.objects.get(name=project_name)
        path = reverse('duat-post', kwargs={'project_name':project.name})
    context = RequestContext(request)
    host = request.META.get('HTTP_HOST')
    context['project'] = project
    context['host'] = host
    context['submit_url'] = path
    context['csrf_token_name'] =settings.CSRF_COOKIE_NAME
    return HttpResponse(render_to_response(filename, context_instance=context),
                        mimetype = 'application/javascript')



from django.contrib import admin
from django.core.urlresolvers import reverse
from django import forms
from django.conf import settings

from duat.models import Feedback, Project


class FeedbackModelForm( forms.ModelForm ):
    comment = forms.CharField( widget=forms.Textarea )

    class Meta:
        model = Feedback


class FeedbackAdmin(admin.ModelAdmin):
    fieldsets = [
        (None,               {'fields': ['project','comment','screenshot']}),
        ('Details', {'fields': ['page','referrer','user_agent','created_date']}),
    ]

    readonly_fields = ('created_date','screenshot','page','referrer','user_agent')
    admin_readonly_fields = ('created_date','project','screenshot','page','referrer','user_agent')
    list_display = ('id_render', 'project', 'comment', 'created_date', 'screenshot')
    _superuser_filter = ('project','created_date') # filters for superuser
    _admin_list_filter = ('created_date',) # filters visible to project admin only
    list_filter = _superuser_filter
    search_fields = ['comment']
    list_per_page = 5
    form = FeedbackModelForm
    
    
    def id_render(self, obj):
        """
        Increase reability of id column
        """
        url = reverse('admin:%s_%s_change' %(obj._meta.app_label,  obj._meta.module_name),  args=[obj.id] )
        return ('<a href="%s">Feedback #%s</a>' % (url, obj.id))
    id_render.allow_tags = True

    def screenshot(self, obj):
        """
        Renders a clickable screenshot with links to the full image or html view
        """
        page_url = reverse('view',kwargs={'project_name':obj.project.name,'id':obj.id})

        PHANTOMJS_EXECUTABLE = getattr(settings, 'PHANTOMJS_EXECUTABLE', '')
        if PHANTOMJS_EXECUTABLE:
            url_image = settings.STATIC_URL + 'duat/screenshots/' + str(obj.id) + '.jpg'
            return ('<a href="%s" style="height:200px;width:200px;'
                    'background-image:url(%s);'
                    'background-repeat:no-repeat;'
                    'display:block;'
                    'background-size: 200px auto;">&nbsp;</a>'
                    'View: <a href="%s">Image</a> | <a href="%s">Page</a>' % 
                    (url_image, url_image, url_image, page_url))
        else:
            # No PhantomJS setup
            return ('View: <a href="%s">Page</a>' %  (page_url))
    screenshot.allow_tags = True

    def has_add_permission(self, request):
        """
        Disable add permission from admin interface
        """
        return False

    def get_readonly_fields(self, request, obj=None):
        """
        Only allow superuser to edit fields
        """
        if not request.user.is_superuser and request.user.has_perm('duat.readonly_feedback'):
            list = [f.name for f in self.model._meta.fields]
            list.append('screenshot')
            return list
        return self.readonly_fields

    def queryset(self, request):
        """
        Limit Feedback to those that belong to the project the request's user is admin of.
        """
        qs = super(FeedbackAdmin, self).queryset(request)
        if request.user.is_superuser:
            # It is mine, all mine. Just return everything.
            return qs
        # Now we just add an extra filter on the queryset and
        # we're done.
        return qs.filter(project__admin=request.user)


    def changelist_view(self, request, extra_context=None):
        """
        Change filter options based on user type
        """
        if request.user.is_superuser:
            self.list_filter = self._superuser_filter
        else:
            self.list_filter = self._admin_list_filter
        return super(FeedbackAdmin, self).changelist_view(request, extra_context)


class ProjectAdmin(admin.ModelAdmin):
    list_display = ('project_name','admin')
    readonly_fields = ('jslink',)

    host = None

    def project_name(self, project):
        return project.name

    def jslink(self, project):
        """
        Provide a link to the embeddable javascript file
        """
        url = reverse('duat-feedback',kwargs={'project_name':project.name})
        return '<script src="%s%s"></script>' % (ProjectAdmin.host, url)
    jslink.short_description = 'Embed link'

    def get_readonly_fields(self, request, obj=None):
        if not ProjectAdmin.host:
            # set the class-level host here was we have a request object
            if request.is_secure():
                ProjectAdmin.host = 'https://' + request.get_host()
            else:
                ProjectAdmin.host = 'http://' + request.get_host()
        if obj: # Editing
            return self.readonly_fields
        return ()

admin.site.register(Project, ProjectAdmin)
admin.site.register(Feedback, FeedbackAdmin)



import os

from django.db import models
from django.core.files.base import ContentFile
from django.conf import settings
from django.db.models.signals import pre_delete
from django.dispatch.dispatcher import receiver

try:
    # Django 1.5
    from django.utils.text import slugify
except:
    # Older versions
    from django.template.defaultfilters import slugify


class Project(models.Model):
    name = models.CharField(max_length=200,
                            unique=True)
    admin = models.ForeignKey(settings.AUTH_USER_MODEL)
    notify_admin = models.BooleanField(default=False)
    
    def save(self, *args, **kwargs):
        self.name = slugify(self.name)
        super(Project, self).save(*args, **kwargs) # Call the "real" save() method.
        
    def __unicode__(self):
        return u'%s' % (self.name)


class Feedback(models.Model):
    project = models.ForeignKey(Project)
    page = models.CharField(max_length=2000)
    referrer = models.CharField(max_length=2000)
    user_agent = models.CharField(max_length=2000)
    comment = models.CharField(max_length=2000, null=True)
    html = models.TextField(null=True)
    image = models.TextField(null=True)
    created_date = models.DateTimeField(auto_now = True)
    
    def __unicode__(self):
        return u'Feedback #%s (%s)' % (self.id, self.project.name)
    
    class Meta:
        ordering = ('-created_date',)
        permissions = (
            ('readonly_feedback','Readonly Feedback'),
        )

@receiver(pre_delete, sender=Feedback)
def _feedback_delete(sender, instance, **kwargs):
    """ Delete auto-generated screenshot when model instance is deleted """
    screenshot = "%s/%s.jpg" % (os.path.join(settings.STATIC_ROOT, 'screenshots'),
                                 instance.id)
    try:
        os.remove(screenshot)
    except OSError:
        pass



import os
from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-duat',
    version='0.2',
    packages=['duat'],
    include_package_data=True,
    license='MIT License',
    description='A lightweight Django User Acceptance Testing feedback tool.',
    long_description=README,
    author='Chris Bailey',
    author_email="chris.p.bailey@gmail.com",
    url='https://github.com/chrispbailey/duat/',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Topic :: Software Development :: Libraries :: Python Modules",
        'Topic :: Software Development :: Testing',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
